from __future__ import annotations

import json
import re
from datetime import datetime
from typing import Any, Dict, cast

from petisco.base.domain.model.uuid import Uuid
from petisco.base.domain.model.value_object import ValueObject

TIME_FORMAT = "%Y-%m-%d %H:%M:%S.%f"


def get_version(config: dict[str, Any] | None) -> int:
    version = getattr(config, "version", 1) if config else 1
    return version


def get_message_name(namespace: dict[str, Any]) -> str:
    message_name = namespace.get("__qualname__", "message")

    if ".<locals>." in message_name:
        message_name = message_name.split(".<locals>.")[-1]
    message_name = message_name.replace("Legacy", "")
    return re.sub(r"(?<!^)(?=[A-Z])", "_", message_name).lower().replace("_", ".")


class MetaMessage(type):
    def __new__(
        mcs, name: str, bases: tuple[Any], namespace: dict[str, Any]
    ) -> MetaMessage:
        config = namespace.get("Config")

        namespace["_message_version"] = get_version(config)
        namespace["_message_name"] = get_message_name(namespace)
        namespace["_message_attributes"] = {}
        namespace["_message_meta"] = {}

        return super().__new__(mcs, name, bases, namespace)


class LegacyMessage(metaclass=MetaMessage):
    _message_id: Uuid
    _message_name: str
    _message_version: int
    _message_occurred_on: datetime
    _message_attributes: dict[str, Any]
    _message_meta: dict[str, Any]
    _message_type: str

    def __init__(self, **data: Any) -> None:
        self._message_attributes = {}
        self._message_meta = {}
        if not hasattr(self, "_message_type"):
            self._message_type = "message"
        self._set_data(**data)

    def _set_data(self, **kwargs: dict[str, Any] | None) -> None:
        if kwargs:
            self._message_id = (
                Uuid.from_value(kwargs.get("id")) if kwargs.get("id") else Uuid.v4()
            )
            self._message_name = str(kwargs.get("type"))
            self._message_version = int(kwargs.get("version", 1))
            self._message_occurred_on = (
                datetime.strptime(str(kwargs.get("occurred_on")), TIME_FORMAT)
                if kwargs.get("occurred_on")
                else datetime.now()
            )
            self._message_attributes = cast(Dict[str, Any], kwargs.get("attributes"))
            self._message_meta = cast(Dict[str, Any], kwargs.get("meta"))
            self._message_type = str(kwargs.get("type_message", self._message_type))

            if self._message_attributes:
                for key, value in self._message_attributes.items():
                    setattr(self, key, value)
        else:
            self._message_id = Uuid.v4()
            self._message_occurred_on = datetime.utcnow()

    def _set_attributes(self, **data: Any) -> None:
        if self._message_attributes is None:
            self._message_attributes = {}
        for k in data:
            self._message_attributes[k] = data[k]
            setattr(self, k, data[k])

    def add_meta(self, meta: dict[str, Any]) -> None:
        self._message_meta = meta

    def update_meta(self, meta: dict[str, Any]) -> LegacyMessage:
        if not meta:
            return self

        if not isinstance(meta, Dict):
            raise TypeError("Message.update_meta() expect a dict")
        if hasattr(self, "_message_meta"):
            self._message_meta = {**self._message_meta, **meta}
        else:
            self._message_meta = meta
        return self

    @staticmethod
    def from_dict(message_data: dict[str, Any]) -> LegacyMessage:
        data = cast(Dict[str, Any], message_data.get("data"))
        return LegacyMessage(**data)

    @staticmethod
    def from_json(message_json: str | bytes) -> LegacyMessage:
        event_dict = json.loads(message_json)
        return LegacyMessage.from_dict(event_dict)

    @classmethod
    def from_format(
        cls,
        formatted_message: dict[str, Any] | str | bytes,
        target_type: type | None = None,
    ) -> LegacyMessage:
        if not isinstance(formatted_message, dict):
            formatted_message = json.loads(formatted_message)
        return cls.from_dict(formatted_message)

    def _get_serialized_attributes(self) -> dict[str, Any]:
        attributes = {}
        for key, attribute in self._message_attributes.items():
            serialized_value = attribute
            if isinstance(attribute, ValueObject):
                serialized_value = attribute.value
            if isinstance(attribute, datetime):
                serialized_value = attribute.strftime(TIME_FORMAT)
            attributes[key] = serialized_value
        return attributes

    def dict(self) -> dict[str, dict[str, Any]]:
        data = {
            "data": {
                "id": self._message_id.value,
                "type": self._message_name,
                "type_message": self._message_type,
                "version": self._message_version,
                "occurred_on": self._message_occurred_on.strftime(TIME_FORMAT),
                "attributes": self._get_serialized_attributes(),
                "meta": self._message_meta,
            }
        }
        return data

    def model_dump(self) -> dict[str, dict[str, Any]]:
        return self.dict()

    def format(self) -> dict[str, dict[str, Any]]:
        # To improve migration
        return self.dict()

    def format_json(self) -> str:
        # To improve migration
        return json.dumps(self.format())

    def to_str(self, class_name: str = "Message", type: str = "message") -> str:
        return f"{class_name} [{self._message_id.value} ({type}), {self._message_name} (v{self._message_version}), {self._message_occurred_on}, attributes={self._message_attributes}]"

    def json(self) -> str:
        return json.dumps(self.dict())

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, LegacyMessage):
            return False
        return self.dict() == other.dict()

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return self.to_str()

    def get_message_id(self) -> Uuid:
        return self._message_id

    def get_message_name(self) -> str:
        return self._message_name

    def get_message_version(self) -> int:
        return self._message_version

    def get_message_occurred_on(self) -> datetime:
        return self._message_occurred_on

    def get_message_attributes(self) -> dict[str, Any]:
        return self._message_attributes

    def get_message_meta(self) -> dict[str, Any]:
        return self._message_meta

    def get_message_type(self) -> str:
        return self._message_type



from __future__ import annotations

import json
from typing import Any, TypeVar

from petisco.base.domain.message.legacy.legacy_message import LegacyMessage

T = TypeVar("T", bound="LegacyCommand")


class LegacyCommand(LegacyMessage):
    """
    A base class to model your Command.
    An operation that effects some change to the system (for example, modify the persistence).
    An operation that intentionally creates a side effect.
    """

    def __init__(self, **data: Any) -> None:
        self._set_data()
        self._set_attributes(**data)
        self._message_type = "command"  # noqa

    @staticmethod
    def from_dict(
        message_data: dict[str, Any], target_type: type[T] | None = None
    ) -> T:
        target_type = LegacyCommand if target_type is None else target_type
        data = message_data.get("data")
        command = target_type()
        command._set_data(**data)  # type: ignore
        return command

    @staticmethod
    def from_json(message_json: str | bytes, target_type: type[T] | None = None) -> T:
        event_dict = json.loads(message_json)
        return LegacyCommand.from_dict(event_dict, target_type)

    def __repr__(self) -> str:
        return self.to_str(class_name="Command", type="command")



from __future__ import annotations

from contextlib import asynccontextmanager
from typing import Callable, ContextManager

from elasticsearch import AsyncElasticsearch
from loguru import logger

from petisco.base.domain.persistence.async_database import AsyncDatabase
from petisco.extra.elastic.elastic_connection import ElasticConnection


def async_elastic_session_scope_provider(session) -> Callable:
    from elasticsearch import RequestError

    @asynccontextmanager
    async def session_scope() -> ContextManager[AsyncElasticDatabase]:
        try:
            yield session
            await session.close()
        except ConnectionRefusedError as e:
            logger.error(e)
            raise e
        except RequestError as e:
            logger.error(e)
            raise e
        except Exception as e:
            logger.error(e)
            raise e

    return session_scope


AsyncElasticSessionScope = Callable[..., ContextManager[AsyncElasticsearch]]


class AsyncElasticDatabase(AsyncDatabase):
    session: AsyncElasticsearch | None = None

    @staticmethod
    def local_connection_checker() -> AsyncElasticDatabase:
        return AsyncElasticDatabase(
            name="test", connection=ElasticConnection.create_local()
        )

    def __init__(self, alias: str, connection: ElasticConnection) -> None:
        if not connection or not isinstance(connection, ElasticConnection):
            raise ConnectionError(
                "ElasticDatabase needs a valid ElasticConnection connection"
            )
        self.connection = connection
        super().__init__(alias)

    def initialize(self) -> None:
        self.session = AsyncElasticsearch(
            self.connection.to_elastic_format(), http_auth=self.connection.http_auth
        )

    def delete(self) -> None:
        pass

    def clear_data(self) -> None:
        pass

    async def is_available(self) -> bool:
        try:
            return await self.session.ping()
        except Exception:  # noqa E722
            return False

    def get_session_scope(self) -> AsyncElasticSessionScope:
        return async_elastic_session_scope_provider(self.session)



from typing import Any

from petisco.base.domain.message.legacy.use_legacy_implementation import (
    USE_LEGACY_IMPLEMENTATION,
)
from petisco.base.domain.message.message import Message


class Command(Message):
    def model_post_init(self, __context: Any) -> None:
        self._message_type = "command"
        super().model_post_init(__context)


if USE_LEGACY_IMPLEMENTATION is True:
    from petisco.base.domain.message.legacy.legacy_command import LegacyCommand  # noqa

    Command = LegacyCommand  # noqa



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, TypeVar

from petisco.base.domain.persistence.async_database import AsyncDatabase
from petisco.base.domain.persistence.database import Database

T = TypeVar("T")


def get_key(base_type: type[T], alias: str | None = None) -> str:
    return base_type.__name__ if not alias else alias


@dataclass
class _Databases:
    def __init__(self) -> None:
        self._databases: dict[str, Any] = {}

    def __repr__(self) -> str:
        return f"Databases: {str(self.info())}"

    def info(self) -> dict[str, Any]:
        return {name: database.info() for name, database in self._databases.items()}

    def get_databases(self) -> list[Database]:
        return list(self._databases.values())

    def get_database_names(self) -> list[str]:
        return list(self._databases.keys())

    def add(
        self, database: Database | list[Database], skip_if_exist: bool = False
    ) -> None:
        if isinstance(database, list):
            for database_ in database:
                self.add(database_, skip_if_exist)

        elif isinstance(database, Database):
            key = database.get_key()
            if key in self._databases:
                if skip_if_exist is False:
                    raise NameError(f"Database {key} is already added to Databases")
            else:
                self._databases[key] = database
        else:
            raise TypeError(
                f"Databases.add only accept a Database or a list of Database and your input is {database}"
            )

    def get(self, base_type: type[T], *, alias: str | None = None) -> T:
        key = get_key(base_type, alias)
        if key in self._databases:
            return self._databases.get(key)
        else:
            raise NameError(f"Database {key} does not exist in Databases")

    def remove(
        self,
        base_type: type[T],
        *,
        alias: str | None = None,
        skip_if_not_exist: bool = False,
    ) -> None:
        key = get_key(base_type, alias)

        if key in self._databases:
            self._databases[key].delete()
            del self._databases[key]
        else:
            if skip_if_not_exist is False:
                raise IndexError(f"Database cannot be removed. {key} does not exists")

    def initialize(
        self, initialization_arguments: dict[str, dict[str, Any]] | None = None
    ) -> None:
        for database in self._databases.values():
            if isinstance(database, AsyncDatabase):
                continue
            arguments = (
                initialization_arguments.get(database.alias)
                if initialization_arguments
                else None
            )
            if arguments:
                database.initialize(**arguments)
            else:
                database.initialize()

    async def async_initialize(
        self, initialization_arguments: dict[str, dict[str, Any]] | None = None
    ) -> None:
        for database in self._databases.values():
            if isinstance(database, AsyncDatabase):
                arguments = (
                    initialization_arguments.get(database.alias)
                    if initialization_arguments
                    else None
                )
                if arguments:
                    await database.initialize(**arguments)
                else:
                    await database.initialize()

    def delete(self) -> None:
        for database in self._databases.values():
            database.delete()

    def clear_database(self, base_type: type[T], *, alias: str | None = None) -> None:
        key = get_key(base_type, alias)
        databases_ = self._databases
        if key is not None:
            if key not in self._databases:
                raise IndexError(f"Database cannot clear the data. {key} not exists")
        for database in databases_.values():
            database.clear_data()

    def clear(self) -> None:
        self._databases = {}


databases = _Databases()



from abc import abstractmethod
from typing import Any, Dict

from meiga import AnyResult, NotImplementedMethodError

from petisco.base.application.controller.controller import Controller
from petisco.base.misc.result_mapper import ResultMapper
from petisco.extra.fastapi.controller.fastapi_result_mapper import FastAPIResultMapper


class FastAPIController(Controller):
    @staticmethod
    def get_default_mapper() -> ResultMapper:
        return FastAPIResultMapper.default()

    @staticmethod
    def get_config_mapper(config: Dict[str, Any]) -> ResultMapper:
        return FastAPIResultMapper.from_config(config)

    @abstractmethod
    def execute(self, *args: Any, **kwargs: Any) -> AnyResult:
        return NotImplementedMethodError

    @classmethod
    def responses(cls) -> Dict[str, Any]:
        controller = cls()

        if not hasattr(controller, "Config"):
            return None

        config = controller.Config
        if not hasattr(config, "error_map"):
            return None

        expected_responses = {
            http_error.status_code: {"description": http_error.detail}
            for http_error in config.error_map.values()
        }
        return expected_responses



from petisco.base.domain.message.message import Message


class RabbitMqMessageQueueNameFormatter:
    @staticmethod
    def format(message: Message, exchange_name: str = None) -> str:
        message_name = message.get_message_name().replace(".", "_")
        message_type = (
            message.get_message_type()
            if message.get_message_type() != "domain_event"
            else "event"
        )
        message_format = (
            f"{message.get_message_version()}.{message_type}.{message_name}"
        )
        return f"{exchange_name}.{message_format}" if exchange_name else message_name

    @staticmethod
    def format_retry(message: Message, exchange_name: str = None) -> str:
        queue_name = RabbitMqMessageQueueNameFormatter.format(message, exchange_name)
        return f"retry.{queue_name}"

    @staticmethod
    def format_dead_letter(message: Message, exchange_name: str = None) -> str:
        queue_name = RabbitMqMessageQueueNameFormatter.format(message, exchange_name)
        return f"dead_letter.{queue_name}"



import os
from functools import wraps
from inspect import signature
from typing import Any, Callable, Dict, List, Type, Union

import elasticapm
from loguru import logger
from meiga import Error, Failure
from meiga.on_failure_exception import OnFailureException

from petisco.base.application.middleware.middleware import Middleware
from petisco.base.application.middleware.notifier_middleware import NotifierMiddleware
from petisco.base.application.middleware.print_middleware import PrintMiddleware
from petisco.base.domain.errors.unknown_error import UnknownError
from petisco.base.misc.result_mapper import ResultMapper


def get_middleware_instances(config: Dict[str, Any]) -> List[Middleware]:
    middlewares_instances: List[Middleware] = []

    middlewares_configs = getattr(config, "middlewares", [])
    if not middlewares_configs:
        middlewares_configs = get_middlewares_configuration_from_environment()

    for middlewares_config in middlewares_configs:
        if not isinstance(middlewares_config, Middleware):
            middlewares_instances.append(middlewares_config())
        else:
            middlewares_instances.append(middlewares_config)

    return middlewares_instances


def get_middlewares_configuration_from_environment():
    def gettype(name: str) -> Type[Middleware]:
        lookup_table = {
            "NotifierMiddleware": NotifierMiddleware,
            "PrintMiddleware": PrintMiddleware,
        }
        my_type = lookup_table.get(name)
        if not my_type:
            raise TypeError(
                f"Value {name} in PETISCO_DEFAULT_MIDDLEWARES is not valid. Please, use {list(lookup_table.keys())}"
            )
        return my_type

    default_middlewares_names: Union[str, None] = os.getenv(
        "PETISCO_DEFAULT_MIDDLEWARES"
    )
    if not default_middlewares_names:
        return []
    default_middlewares_names_list = default_middlewares_names.split(",")
    return [
        gettype(middleware_name) for middleware_name in default_middlewares_names_list
    ]


def wrapper(
    execute_func: Callable[..., Any],
    wrapped_class_name: str,
    config: Any,
    mapper: ResultMapper,
) -> Callable[..., Any]:
    @wraps(execute_func)
    def wrapped(*args: Any, **kwargs: Any) -> Any:
        middlewares = get_middleware_instances(config)
        arguments = signature(execute_func).bind(*args, **kwargs).arguments
        arguments.pop("self")

        for middleware in middlewares:
            try:
                middleware.set_data(wrapped_class_name, arguments)
                middleware.before()
            except Exception as exception:
                logger.error(f"Error in the {wrapped_class_name} middlewares (before).")
                logger.exception(exception)

        try:
            result = execute_func(*args, **kwargs)
        except OnFailureException as exc:
            result = exc.result
        except Error as error:
            result = Failure(error)
        except Exception as exception:
            unknown_error = UnknownError.from_exception(
                exception=exception,
                arguments=arguments if len(arguments) > 0 else args,
                class_name=wrapped_class_name,
            )
            result = Failure(unknown_error)
            client = elasticapm.get_client()
            if client:
                client.capture_exception()

        for middleware in middlewares:
            if result:
                try:
                    middleware.after(result)
                except Exception as exception:
                    logger.error(
                        f"Error in the {wrapped_class_name} middlewares (after)."
                    )
                    logger.exception(exception)

        try:
            result.set_transformer(mapper.map)
        except AttributeError:  # noqa
            raise TypeError(
                f"Controller Error: Return value `{result}` ({type(result)}) must be a `meiga.Result` to "
                f"transform values to success and failure handlers."
            )
        return result

    return wrapped



from __future__ import annotations

import json
from typing import Any, TypeVar

from petisco.base.domain.message.legacy.legacy_message import LegacyMessage

T = TypeVar("T", bound="LegacyDomainEvent")


class LegacyDomainEvent(LegacyMessage):
    """
    A base class to model your domain events.
    Define your Domain Events to express what happened in your domain.
    """

    def __init__(self, **data: Any) -> None:
        super().__init__()
        self._set_attributes(**data)
        self._message_type = "domain_event"

    @staticmethod
    def from_dict(
        message_data: dict[str, Any],
        target_type: type[T] | None = None,
    ) -> T:
        target_type = LegacyDomainEvent if target_type is None else target_type
        data = message_data.get("data")
        domain_event = target_type()
        domain_event._set_data(**data)  # type: ignore
        return domain_event

    @staticmethod
    def from_json(
        message_json: str | bytes,
        target_type: type[T] | None = None,
    ) -> T:
        event_dict = json.loads(message_json)
        return LegacyDomainEvent.from_dict(event_dict, target_type)

    def __repr__(self) -> str:
        return self.to_str(class_name="DomainEvent", type="domain_event")



from __future__ import annotations

import os

from petisco.base.application.application_configurer import ApplicationConfigurer
from petisco.base.application.dependency_injection.container import Container
from petisco.base.domain.message.message_configurer import MessageConfigurer
from petisco.base.domain.message.message_consumer import MessageConsumer
from petisco.base.domain.message.message_subscriber import MessageSubscriber

MAX_RETRIES = 5

CLEAR_SUBSCRIBER_BEFORE: bool = os.getenv(
    "PETISCO_RABBITMQ_CONFIGURER_CLEAR_SUBSCRIBER_BEFORE", "false"
).lower() in ["true", "1", "yes", "on"]
CLEAR_STORE_BEFORE: bool = os.getenv(
    "PETISCO_RABBITMQ_CONFIGURER_CLEAR_STORE_BEFORE", "false"
).lower() in ["true", "1", "yes", "on"]


class RabbitMqConfigurer(ApplicationConfigurer):
    def __init__(
        self,
        subscribers: list[type[MessageSubscriber]],
        execute_after_dependencies: bool = True,
        start_consuming: bool = True,
        alias: str = None,
    ):
        """
        Initializes an instance of RabbitMqConfigurer.

        Args:
            subscribers (list[MessageSubscriber]): A list of MessageSubscriber objects representing the message subscribers.
            execute_after_dependencies (bool, optional): Flag indicating whether to execute after dependencies. Defaults to True.
            start_consuming (bool, optional): Flag indicating whether to start consuming messages. Defaults to True.
            alias (str, optional): Alias for the MessageConfigurer and MessageConsumer (Container.get(MessageConfigurer|MessageConsumer, alias=self.alias)). Defaults to None.
        """
        self.subscribers = subscribers
        self.start_consuming = start_consuming
        self.alias = alias
        super().__init__(execute_after_dependencies)

    def execute(self, testing: bool = False) -> None:
        configurer = Container.get(MessageConfigurer, alias=self.alias)
        configurer.configure_subscribers(
            self.subscribers,
            clear_subscriber_before=CLEAR_SUBSCRIBER_BEFORE,
            clear_store_before=CLEAR_STORE_BEFORE,
        )

        if self.start_consuming:
            consumer = Container.get(MessageConsumer, alias=self.alias)
            consumer.add_subscribers(self.subscribers)
            consumer.start()



from petisco.base.domain.persistence.databases import databases
from petisco.extra.elastic.elastic_database import ElasticDatabase


def elastic_is_running_locally() -> bool:
    try:
        database = ElasticDatabase.local_connection_checker()
        databases.add(database)
        databases.initialize()
        is_running_locally = databases.get(ElasticDatabase).is_available()
    except:  # noqa: E722
        is_running_locally = False
    databases.remove(ElasticDatabase, skip_if_not_exist=True)
    return is_running_locally



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Callable, ContextManager, Generic, TypeVar

T = TypeVar("T")


class Database(Generic[T], ABC):
    alias: str | None = None

    def __init__(self, alias: str | None = None):
        self.alias = alias

    def get_key(self) -> str:
        return type(self).__name__ if not self.alias else self.alias

    @classmethod
    def info(cls) -> dict[str, Any]:
        return {"type": cls.__name__, "alias": cls.alias}

    @abstractmethod
    def initialize(self, *args: Any, **kwargs: Any) -> None:
        raise NotImplementedError

    @abstractmethod
    def delete(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def clear_data(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def is_available(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def get_session_scope(self) -> Callable[..., ContextManager[T]]:
        raise NotImplementedError
